/* 
 * serial.c - routines need by ccupload.c and ccdownload.c
 * 
 * Copyright (C) 2003 Chris Ruehl <ruehlc@wiesbaden.vistec.net>
 * original code-parts from Diman Todorov
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * ----------------------------------------------------------- 
 *  who | date | Comment
 * -----+------+----------------------------------------------
 *  cr   031007  Basic-Bytes length check changed validate to
 *  		 8KB 
 * 
 */

#include <stdio.h>   /* Standard input/output definitions */
#include <errno.h>   /* Error number definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <stdlib.h>  /* atoi need it */
#include <netinet/in.h> /* proto htons */

#define WAITFORCC 200000
#define READMAX 256

int ccmicro = 0;

int sendit(int sfd, char *it ,size_t size) {
  int cnt;
  unsigned char buf[READMAX];
  
  memset(buf,0,READMAX*sizeof(char));
  cnt=write(sfd, it , size);
  if ( cnt <= 0 ) {
	  perror("serial write");
      return -1;
  }
  usleep(10000);
  read(sfd,&buf,READMAX);
  /* fprintf(stderr,"(read %i %i %i %i %i )\n", 
    			buf[0],buf[1],buf[2],buf[3],buf[4]); */
 return 0;
}

int file_avail (char *this) {
    int fd; 
    
    fd = open(this, O_RDONLY);
    if ( fd == -1 ) {
        perror(this);       	
	exit(-1);
    }
    close (fd);
    return 0;
}

/*
 * open a serial port
 * 9600 8N1 Software-flow-controled
 */
int open_port(char *device)
{
    int fd; 
    struct termios options;
    
    fd = open(device, O_RDWR | O_NOCTTY);
    if ( fd == -1 ) {
        perror("open_port");       	
	return -1;
    }
	
    memset((char *)&options,0,sizeof(options));
    
    tcgetattr(fd, &options);
    
    /* 8n1 ignore modem-lines software flow-control */
    options.c_cflag = CS8 | CLOCAL | CREAD;
    options.c_iflag = IGNPAR;
    options.c_oflag = 0;
    options.c_lflag = 0;

    /* set io-seed to 9600 */
    cfsetospeed(&options,B9600);
    cfsetispeed(&options,B9600);

    /* flush io-buffers */
    tcflush(fd, TCIOFLUSH);
    tcsetattr(fd, TCSANOW, &options);

    return (fd);
}

int initmc(int fd)
{
#define INITBS 256
    int i = 0;
    char sendc;
    char recvc[INITBS];
    char chout = 'x';

    memset(recvc,0,INITBS*sizeof(char));
    
    sendc = 0x01;  //Turn On 
    
    fprintf(stderr,"Initialize C-Control\n");
    write(fd, &sendc, sizeof(char)); 
    usleep(WAITFORCC);

    while ( i < INITBS ) //get version string or recv max bytes
    {
       read(fd, &chout, sizeof(chout));
       recvc[i++] = chout;
       if(chout == '\r' ) {  // get a break
           break;
       }
    }
    if ( strcmp(recvc, "CCTRL-BASIC Version 2.0 (20.12.03)\r")==0) ccmicro = 1;
    if(! strncmp(recvc, "CCTRL-BASIC Version", 19) ) {
	fprintf(stderr,"Current C-Control:\n%s\n",recvc);
	return 0;
    }
    else { 
	fprintf(stderr,"failed no C-Control found\n");
        return -1;
    }
}

int sendprog(char *filen, int sfd , char sflag)
{
#define SENDBS 256
    int i,val;
    FILE *stream;
    unsigned short length;
    char hibyte,lobyte;
    char buf[SENDBS];
    char ch, *code , *pos;
    int statistic[2]={0 ,0 };
    
    stream = fopen(filen,"r"); 
    if ( stream == NULL ) {
	perror(filen);
	return -1;
    }
    
    /* get header */
    fgets(buf,(SENDBS*sizeof(char)),stream);
    
    if(strcmp(buf, "CCTRL-BASIC\n") != 0) {
    	fprintf(stderr,"Wrong file format.\n");    
	fclose(stream);
	return -1;
    }
    
    /* get length */
    fgets(buf,(SENDBS*sizeof(char)),stream);
    length =(unsigned short) atoi(buf);
    if ( length == 0 || length > (1024*8) ) {
       fprintf(stderr,"%s , abort.\n",length==0?"Null Basic-Bytes":
	       "Maxlength of 8k for Basic-Bytes exceeded" );    
       fprintf(stderr,"Length: %i\n",length);    
       fclose(stream);
       return -1;
    }
    statistic[0]=length;
    fprintf(stderr,"Basic-Bytes %i\n",length);
    /* prepare code memory allocation */
    if (  (code =(char *) malloc(length *sizeof(char)) ) == NULL ) {
       fprintf(stderr,"Memory allocation error, abort.\n");
       fclose(stream);
       return -1;
    }
    memset ( (char *) code , 0, length *sizeof(char));
    
    pos=code;
    i=0;
    while ( code+length != pos && (val = fgetc(stream)) != EOF ) {
	switch ( val ) {
	    case '\r':
		;; // nop
		break;
	    case '\n':
		;; // nop
		break;
	    case ' ' :
	    buf[i]='\0';
	    *pos++=(char)atoi(buf);	
	    i=0;
	    break;
	    default:
	    buf[i++]=(char)val;
	    break;
	}
    }
    
    /* change to network byte order */
    hibyte  = (char)((length>>8)&0x00ff);
    lobyte = (char)((length)&0x00ff );
    /* fprintf(stderr,"(hi=%i,lo=%i)\n",hibyte,lobyte); */
    
    //prepare mc for eprom programming
    tcflush(sfd, TCIOFLUSH);
    ch = 0x02;
    write(sfd, &ch , sizeof(char));
    usleep(WAITFORCC);

    //send prog length    
    write(sfd, &hibyte , sizeof(char));
    /* -- undocumented .. if hi = 0 no echo is returned */
    if ( hibyte || ccmicro ) {
       read(sfd, &buf, sizeof(char));
    }
    write(sfd, &lobyte , sizeof(char));
    read(sfd, &buf, sizeof(char));

    //send program
    for ( i = 0 , pos=code ; i < length ;i ++) {
     if ( sendit(sfd,pos++,sizeof(char))  < 0 ) {
	 free(code);
	 return -1;
     }
     else{
	 fprintf(stderr,"%i/%i\r",i+1,length);
     }
    }
    fprintf(stderr,"\n");

    /* --------------------------*/
    /* -- send system bytes ? -- */
    if ( sflag > 0 ) {
    buf[0]=0;  /* eat all \n + white-space-lines */
    while ( strlen(buf) < 2 ) {  
     fgets(buf,(SENDBS*sizeof(char)),stream);
    }
    length = (unsigned short)atoi(buf);
    if (  length > 255 ) {
       fprintf(stderr,"System Bytes: invalid length (%i) detected, abort.\n",
	       length);    
       fclose(stream);
       return -1;
    }
    
    if ( length == 0 ) {
	fprintf(stderr,"nothing for System-Bytes to do :-)\n");
    }
    else  {
    fprintf(stderr,"System-Bytes %i\n",length);
    statistic[1]=length;
    free(code);
    /* prepare code memory allocation */
    if (  (code =(char *) malloc(length *sizeof(char)) ) == NULL ) {
       fprintf(stderr,"Memory allocation error, abort.\n");
       fclose(stream);
       return -1;
    }
    memset ( (char *) code , 0, length *sizeof(char));
    
    pos=code;
    i=0;
    while ( code+length != pos && (val = fgetc(stream)) != EOF ) {
	switch ( val ) {
	    case '\r':
		;; // nop
		break;
	    case '\n':
		;; // nop
		break;
	    case ' ' :
	    buf[i]='\0';
	    *pos++=(char)atoi(buf);	
	    i=0;
	    break;
	    default:
	    buf[i++]=(char)val;
	    break;
	}
    }
    
    /* change to network byte order */
    
    //prepare mc for eprom programming
    tcflush(sfd, TCIOFLUSH);
    ch = 0x08;
    write(sfd, &ch , sizeof(char));
    usleep(WAITFORCC);
    //send prog length    
    ch = (unsigned char) length;
    write(sfd, &ch , sizeof(char));
    usleep(20000); 
    //send program
    for ( i = 0 , pos=code ; i < length ;i ++) {
     if ( sendit(sfd,pos++,sizeof(char))  < 0 ) {
	 fclose(stream);
	 free(code);
	 return -1;
     }
     else{
	 fprintf(stderr,"%i/%i\r",i+1,length);
     }
    }
    fprintf(stderr,"\n");
    }		
    }
    fprintf(stderr,"Send %i Basic-Bytes , %i System-Bytes \n",
	    statistic[0],statistic[1]);
    /* free allocted memory */
    free(code);
    fclose(stream);
    tcflush(sfd, TCIOFLUSH);
    return 0;

}


